@@ -75,8 +75,15 @@ class Defect(MSONable, metaclass=ABCMeta):
         )
         self.user_charges = user_charges if user_charges else []
         if oxi_state is None:
-            # TODO this step might take time so wrap it in a timer
-            self.structure.add_oxidation_state_by_guess()
+            # Try to use the reduced cell first since oxidation state assignment
+            # scales poorly with systems size.
+            try:
+                self.structure.add_oxidation_state_by_guess(max_sites=-1)
+                # check oxi_states assigned and not all zero
+                if all([specie.oxi_state == 0 for specie in self.structure.species]):
+                    self.structure.add_oxidation_state_by_guess()
+            except:  # pragma: no cover
+                self.structure.add_oxidation_state_by_guess()
             self.oxi_state = self._guess_oxi_state()
         else:
             self.oxi_state = oxi_state
@@ -680,6 +687,20 @@ def get_element(sp_el: Species | Element) -> Element:
     return sp_el
 
 
+def get_vacancy(structure: Structure, isite: int, **kwargs) -> Vacancy:
+    """Get a vacancy defect from a structure and site index.
+
+    Convenience function for creating a Vacancy object quickly.
+
+    Args:
+        structure: The structure to create the vacancy in.
+        isite: The site index of the vacancy.
+        **kwargs: Keyword arguments to pass to Vacancy constructor.
+    """
+    site = structure[isite]
+    return Vacancy(structure=structure, site=site, **kwargs)
+
+
 # TODO: matching defect complexes might be done with some kind of CoM site to fix the periodicity
 # Get this by taking the periodic average of all the provided sites.
 # class DefectComplex(DummySpecies):
