@@ -52,6 +52,9 @@ class DefectEntry(MSONable):
             The fractional coordinates of the defect in the supercell.
             If None, structures attributes of the locpot file will be used to
             automatically determine the defect location.
+        bulk_entry:
+            The ComputedEntry for the bulk. If this is provided, the energy difference
+            can be calculated from this automatically.
         corrections:
             A dictionary of corrections to the energy.
         corrections_metadata:
@@ -59,6 +62,7 @@ class DefectEntry(MSONable):
             about how the corrections were calculated.  These should are only used
             for debugging and plotting purposes.
             PLEASE DO NOT USE THIS AS A CONTAINER FOR IMPORTANT DATA.
+
     """
 
     defect: Defect
@@ -67,10 +71,13 @@ class DefectEntry(MSONable):
     corrections: dict[str, float] = field(default_factory=dict)
     corrections_metadata: dict[str, Any] = field(default_factory=dict)
     sc_defect_frac_coords: tuple[float, float, float] | None = None
+    bulk_entry: ComputedEntry | None = None
+    entry_id: str | None = None
 
     def __post_init__(self) -> None:
         """Post-initialization."""
         self.charge_state = int(self.charge_state)
+        self.entry_id = self.entry_id or self.sc_entry.entry_id
 
     def get_freysoldt_correction(
         self,
@@ -125,6 +132,7 @@ class DefectEntry(MSONable):
                 defect_structure=defect_struct,
                 base_structure=bulk_struct,
             )
+            self.sc_defect_frac_coords = defect_fpos
         else:
             defect_fpos = self.sc_defect_frac_coords
 
@@ -150,14 +158,31 @@ class DefectEntry(MSONable):
         """The energy of the defect entry with all corrections applied."""
         return self.sc_entry.energy + self.corrections["freysoldt"]
 
+    def get_ediff(self) -> float | None:
+        """Get the energy difference between the defect and the bulk (including finite-size correction)."""
+        if self.bulk_entry:
+            return self.corrected_energy - self.bulk_entry.energy
+        else:
+            return None
+
+    def get_summary_dict(self) -> dict:
+        """Get a summary dictionary for the defect entry."""
+        corrections_d = {f"correction_{k}": v for k, v in self.corrections.items()}
+        res = {
+            "name": self.defect.name,
+            "charge_state": self.charge_state,
+            "bulk_total_energy": self.bulk_entry.energy if self.bulk_entry else None,
+            "defect_total_energy": self.sc_entry.energy,
+        }
+        res.update(corrections_d)
+        return res
+
 
 @dataclass
 class FormationEnergyDiagram(MSONable):
     """Formation energy.
 
     Attributes:
-        bulk_entry:
-            The bulk computed entry to get the total energy of the bulk supercell.
         defect_entries:
             The list of defect entries for the different charge states.
             The finite-size correction should already be applied to these.
@@ -170,6 +195,12 @@ class FormationEnergyDiagram(MSONable):
             The VBM of the bulk crystal.
         band_gap:
             The band gap of the bulk crystal.
+        bulk_entry:
+            The bulk computed entry to get the total energy of the bulk supercell.
+            This is only used in case where the bulk entry data is not provided by
+            the individual defect entries themselves. Default is None.
+            The data from the `bulk_entry` attached to each `defect_entry` will be
+            preferrentially used if it is available.
         inc_inf_values:
             If False these boundary points at infinity are ignored when we look at the
             chemical potential limits.
@@ -182,11 +213,11 @@ class FormationEnergyDiagram(MSONable):
 
     """
 
-    bulk_entry: ComputedStructureEntry
     defect_entries: List[DefectEntry]
     pd_entries: list[ComputedEntry]
     vbm: float
     band_gap: Optional[float] = None
+    bulk_entry: Optional[ComputedStructureEntry] = None
     inc_inf_values: bool = False
 
     def __post_init__(self):
@@ -202,11 +233,18 @@ class FormationEnergyDiagram(MSONable):
                 "Defects are not of same type! "
                 "Use MultiFormationEnergyDiagram for multiple defect types"
             )
+        # if all of the `DefectEntry` objects have the same `bulk_entry` then `self.bulk_entry` is not needed
+        if any(d.bulk_entry is None for d in self.defect_entries):
+            if self.bulk_entry is None:
+                raise RuntimeError(
+                    "Not all of the `DefectEntry` objects have a `bulk_entry` attribute, you need to provide `bulk_entry` to `FormationEnergyDiagram`"
+                )
 
+        bulk_entry = self.bulk_entry or self.defect_entries[0].bulk_entry
         pd_ = PhaseDiagram(self.pd_entries)
-        entries = pd_.stable_entries | {self.bulk_entry}
+        entries = pd_.stable_entries | {bulk_entry}
         pd_ = PhaseDiagram(entries)
-        self.phase_diagram = ensure_stable_bulk(pd_, self.bulk_entry)
+        self.phase_diagram = ensure_stable_bulk(pd_, bulk_entry)
         entries = []
         for entry in self.phase_diagram.stable_entries:
             d_ = dict(
@@ -220,7 +258,7 @@ class FormationEnergyDiagram(MSONable):
 
         self.chempot_diagram = ChemicalPotentialDiagram(entries)
         chempot_limits = self.chempot_diagram.domains[
-            self.bulk_entry.composition.reduced_formula
+            bulk_entry.composition.reduced_formula
         ]
 
         if self.inc_inf_values:
@@ -239,11 +277,11 @@ class FormationEnergyDiagram(MSONable):
     @classmethod
     def with_atomic_entries(
         cls,
-        bulk_entry: ComputedEntry,
         defect_entries: list[DefectEntry],
         atomic_entries: list[ComputedEntry],
         phase_diagram: PhaseDiagram,
         vbm: float,
+        bulk_entry: ComputedEntry | None = None,
         **kwargs,
     ):
         """Create a FormationEnergyDiagram object using an existing phase diagram.
@@ -257,12 +295,10 @@ class FormationEnergyDiagram(MSONable):
         As long as the atomic phase energies are computed using the same settings as
         the defect supercell calculations, the method used to determine the enthalpy of
         formation of the different competing phases is not important.
-        Then use the an exerimentally corrected ``PhaseDiagram`` object (like the ones you can
+        Then use the an experimentally corrected ``PhaseDiagram`` object (like the ones you can
         obtain from the Materials Project) to calculate the enthalpy of formation.
 
         Args:
-            bulk_entry:
-                The bulk computed entry to get the total energy of the bulk supercell.
             defect_entries:
                 The list of defect entries for the different charge states.
                 The finite-size correction should already be applied to these.
@@ -274,6 +310,8 @@ class FormationEnergyDiagram(MSONable):
                 A separately computed phase diagram.
             vbm:
                 The VBM of the bulk crystal.
+            bulk_entry:
+                The bulk computed entry to get the total energy of the bulk supercell.
             band_gap:
                 The band gap of the bulk crystal.
             inc_inf_values:
@@ -289,7 +327,6 @@ class FormationEnergyDiagram(MSONable):
         adjusted_entries = _get_adjusted_pd_entries(
             phase_diagram=phase_diagram, atomic_entries=atomic_entries
         )
-
         return cls(
             bulk_entry=bulk_entry,
             defect_entries=defect_entries,
@@ -347,7 +384,6 @@ class FormationEnergyDiagram(MSONable):
                 defect_locpot=q_locpot, bulk_locpot=bulk_locpot, dielectric=dielectric
             )
             def_entries.append(q_d_entry)
-
         if vbm is None:
             vr = Vasprun(get_zfile(Path(directory_map["bulk"]), "vasprun.xml"))
             vbm = vr.get_band_structure().get_vbm()["energy"]
@@ -404,11 +440,16 @@ class FormationEnergyDiagram(MSONable):
                 for el, fac in defect_entry.defect.element_changes.items()
             ]
         )
-        formation_en = (
-            defect_entry.corrected_energy
-            - (self.bulk_entry.energy + en_change)
-            + self.vbm * defect_entry.charge_state
-        )
+        ediff = defect_entry.get_ediff()
+        if ediff is not None:
+            formation_en = ediff - en_change + self.vbm * defect_entry.charge_state
+        else:
+            formation_en = (
+                defect_entry.corrected_energy
+                - self.bulk_entry.energy
+                - en_change
+                + self.vbm * defect_entry.charge_state
+            )
         return formation_en
 
     @property
@@ -508,6 +549,15 @@ class FormationEnergyDiagram(MSONable):
             energy=fe, temperature=temperature
         )
 
+    def as_dataframe(self):
+        """Return the formation energy diagram as a pandas dataframe."""
+        from pandas import DataFrame
+
+        defect_entries = self.defect_entries
+        l_ = map(lambda x: x.get_summary_dict(), defect_entries)
+        df = DataFrame(l_)
+        return df
+
 
 @dataclass
 class MultiFormationEnergyDiagram(MSONable):
@@ -595,7 +645,10 @@ class MultiFormationEnergyDiagram(MSONable):
 
 
 def group_defects(defect_entries: list[DefectEntry]):
-    """Group defects by their representation."""
+    """Group defects by their representation.
+
+    TODO: Fix this with `defect_structure` grouping with `StructureMatcher`.
+    """
     sents = sorted(defect_entries, key=lambda x: x.defect.__repr__())
     for k, group in groupby(sents, key=lambda x: x.defect.__repr__()):
         yield k, list(group)
